#include <windows.h>
#include <commctrl.h>
#include <string.h>
#include <stdio.h>
#include "getdir.h"


#define IDC_DIRTREE 101


HTREEITEM AddTreeItem(HWND parent, HTREEITEM parentitem, char *text)
{
  TV_INSERTSTRUCT tvis;
  tvis.hParent = parentitem;
  tvis.hInsertAfter = TVI_SORT;
  tvis.item.mask = TVIF_TEXT;
  tvis.item.pszText = text;
  tvis.item.cchTextMax = strlen(text);
  return (HTREEITEM)SendDlgItemMessage(parent, IDC_DIRTREE, TVM_INSERTITEM, 0, (LPARAM)&tvis);
}


bool DirHasSubdirs(char *dir)
{
  char TempDir[256];
  HANDLE File;
  WIN32_FIND_DATA FileData;

  GetCurrentDirectory(255, TempDir);
  SetCurrentDirectory(dir);

  File = FindFirstFile("*.*", &FileData);
  if (File == INVALID_HANDLE_VALUE) return false;

  bool Done = false;
  while (!Done)
  {
    if (FileData.dwFileAttributes == FILE_ATTRIBUTE_DIRECTORY)
      if (FileData.cFileName[0] != '.')
        { FindClose(File); return true; }

    BOOL errorval = FindNextFile(File, &FileData);
    if (errorval == FALSE || errorval == ERROR_NO_MORE_FILES) Done = true;
  }

  SetCurrentDirectory(TempDir);

  FindClose(File);
  return false;
}


void SetItemExpandable(HWND hwnd, HTREEITEM item)
{
  TV_ITEM tvi;
  tvi.mask = TVIF_CHILDREN;
  tvi.hItem = item;
  tvi.cChildren = 1;
  SendDlgItemMessage(hwnd, IDC_DIRTREE, TVM_SETITEM, 0, (LPARAM)&tvi);
}


void AddDrives(HWND hwnd)
{
  char TempDir[256];
  GetCurrentDirectory(255, TempDir);

  HTREEITEM RootItem;

  int i;
  for (i = 2; i < 26; i++)
  {
    char drivename[4] = " :\\";
    drivename[0] = (char)i + 'A';
    if (GetDriveType(drivename) == DRIVE_FIXED)
    {
      SetCurrentDirectory(drivename);
      drivename[2] = 0;
      RootItem = AddTreeItem(hwnd, NULL, drivename);
      drivename[3] = '\\';

      if (DirHasSubdirs(drivename))
        SetItemExpandable(hwnd, RootItem);
    }
  }

  SetCurrentDirectory(TempDir);
}


void AddItemSubdirs(HWND hwnd, HTREEITEM item, char *dir)
{
  char TempDir[256];
  HANDLE File;
  WIN32_FIND_DATA FileData;

  GetCurrentDirectory(255, TempDir);
  SetCurrentDirectory(dir);

  File = FindFirstFile("*.*", &FileData);
  if (File == INVALID_HANDLE_VALUE) return;

  bool Done = false;
  while (!Done)
  {
    if (FileData.dwFileAttributes == FILE_ATTRIBUTE_DIRECTORY)
    {
      if (FileData.cFileName[0] != '.')
      {
        char AbsoluteDir[256];
        sprintf(AbsoluteDir, "%s%s", dir, FileData.cFileName);

        HTREEITEM hitem = AddTreeItem(hwnd, item, FileData.cFileName);
        if (DirHasSubdirs(AbsoluteDir))
          SetItemExpandable(hwnd, hitem);
      }
    }

    BOOL errorval = FindNextFile(File, &FileData);
    if (errorval == FALSE || errorval == ERROR_NO_MORE_FILES) Done = true;
  }

  FindClose(File);

  SetCurrentDirectory(TempDir);
}


int GetItemString(HWND hwnd, HTREEITEM item, char *string)
{
  TV_ITEM tvi;
  tvi.mask = TVIF_TEXT;
  tvi.hItem = item;
  tvi.pszText = string;
  tvi.cchTextMax = 256;
  SendDlgItemMessage(hwnd, IDC_DIRTREE, TVM_GETITEM, 0, (LPARAM)&tvi);
  return strlen(string);
}


int CountCharsIn(char *string, char chr)
{
  int numchars = 0;

  char *c = string;
  while (*c != 0)
  {
    if (*c == chr) numchars++;
    c++;
  }

  return numchars;
}


void GetAbsoluteDir(HWND hwnd, HTREEITEM item, char *dir)
{
  char TempDir[256];

  int length = GetItemString(hwnd, item, TempDir);
  TempDir[length] = '>';
  TempDir[length + 1] = 0;
  length++;

  HTREEITEM parent = (HTREEITEM)SendDlgItemMessage(hwnd, IDC_DIRTREE, TVM_GETNEXTITEM, TVGN_PARENT, (LPARAM)item);
  while (parent != NULL)
  {
    length += GetItemString(hwnd, parent, TempDir + length);
    TempDir[length]     = '>';
    TempDir[length + 1] = 0;
    length++;
    parent = (HTREEITEM)SendDlgItemMessage(hwnd, IDC_DIRTREE, TVM_GETNEXTITEM, TVGN_PARENT, (LPARAM)parent);
  }

  int numitems = CountCharsIn(TempDir, '>');

  int *OffsetList;
  OffsetList = new int[numitems];
   OffsetList[numitems - 1] = 0;

  char *curtok = strchr(TempDir, '>') + 1;

  int i;
  for (i = numitems - 2; i >= 0; i--)
  {
    OffsetList[i] = curtok - TempDir;
    curtok = strchr(curtok, '>') + 1;
  }

  length = 0;
  for (i = 0; i < numitems; i++)
  {
    char *c;
    c = TempDir + OffsetList[i];
    while (*c != 0 && *c != '>')
    {
      dir[length] = *c;
      c++;
      length++;
    }
    dir[length] = '\\';
    dir[length + 1] = 0;
    length++;
  }

  delete OffsetList;
}



typedef char DIR[256];
bool AddDir(DIR dir);
bool InList(DIR dir);
DIR *DirList;
int NumDirs;


#pragma argsused
static BOOL CALLBACK GetDirDialogProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
  static char *dir;
  static HTREEITEM SelItem = NULL;

  switch (msg)
  {
    case WM_INITDIALOG:
      dir = (char *)lparam;
      DirList = NULL;
      NumDirs = 0;
      AddDrives(hwnd);
      return 1;

    case WM_COMMAND:
      switch (wparam)
      {
        case IDOK:
          if (SelItem == NULL)
            { MessageBox(hwnd, "Please select a directory", "GetDirDialog", MB_OK); return 0; }
          GetAbsoluteDir(hwnd, SelItem, dir);
          EndDialog(hwnd, 1);
          return 1;
        case IDCANCEL: EndDialog(hwnd, 0); return 1;
      }
      return 0;

    case WM_NOTIFY:
      NMHDR *nmh;
      nmh = (NMHDR *)lparam;

      if (nmh->idFrom == IDC_DIRTREE)
      {
        if(nmh->code == TVN_ITEMEXPANDING)
        {
          NM_TREEVIEW *nmtv;
          nmtv = (NM_TREEVIEW *)lparam;

          if (nmtv->action == TVE_EXPAND)
          {
            // if expanding
            char DirString[256];

            TV_ITEM tvi;
            tvi.mask = TVIF_TEXT;
            tvi.hItem = nmtv->itemNew.hItem;
            tvi.pszText = DirString;
            tvi.cchTextMax = 256;
            SendDlgItemMessage(hwnd, IDC_DIRTREE, TVM_GETITEM, 0, (LPARAM)&tvi);

            char AbsoluteDir[256];
            GetAbsoluteDir(hwnd, tvi.hItem, AbsoluteDir);

            if (!InList(AbsoluteDir))
            {
              AddItemSubdirs(hwnd, tvi.hItem, AbsoluteDir);
              AddDir(AbsoluteDir);
            }
            return 1;
          }
        }
        if (nmh->code == TVN_SELCHANGED)
        {
          NM_TREEVIEW *nmtv;
          nmtv = (NM_TREEVIEW *)lparam;
          SelItem = nmtv->itemNew.hItem;
        }
      }
      return 0;

    case WM_CLOSE:
      if (DirList != NULL)
        delete DirList;
      NumDirs = 0;
      return 0;

    default: return 0;
  }
}


int GetDirDialog(HWND hwnd, HINSTANCE hInstance, char *dir)
{
  return DialogBoxParam(hInstance, "GetDirDialog", hwnd, (DLGPROC)GetDirDialogProc, (LPARAM)dir);
}


// The directory list used to keep track of which directories have been expanded in the tree view
// is kept in this set of functions and vars.


bool AddDir(DIR dir)
{
  DIR *OldDir = DirList;

  DirList = new DIR[NumDirs + 1];
  if (DirList == NULL)
    { DirList = OldDir; return false; }

  int i;
  for (i = 0; i < NumDirs; i++)
    memcpy(DirList[i], OldDir[i], 256);
  memcpy(DirList[NumDirs], dir, 256);

  NumDirs++;
  delete OldDir;

  return false;
}


bool InList(DIR dir)
{
  int i;
  for (i = 0; i < NumDirs; i++)
    if (strcmp(DirList[i], dir) == 0)
      return true;
  return false;
}
