#include <windows.h>
#include <stdio.h>
#include "midiplay.h"


PLAYLIST Playlist;


bool Shuffle = false;


bool ToggleShuffle()
{
  if (Shuffle) Shuffle = false;
  else         Shuffle = true;
  return true;
}


bool GetShuffle()
{
  return Shuffle;
}


bool SetShuffle(bool value)
{
  Shuffle = value;
  return true;
}


void UpdatePlaylistDialog(HWND hwnd)
{
  SendDlgItemMessage(hwnd, IDC_PLAYLISTBOX, LB_RESETCONTENT, 0, 0);

  int i;
  for (i = 0; i < Playlist.GetNumFiles(); i++)
  {
    SFILENAME *fn = Playlist.GetValue(i);
    SendDlgItemMessage(hwnd, IDC_PLAYLISTBOX, LB_ADDSTRING, 0, (LPARAM)fn);
  }
}


void AddPlaylistDirectory(char *filter, char *dir, bool Recurse)
{
  char TempDir[256];
  GetCurrentDirectory(255, TempDir);
  SetCurrentDirectory(dir);

  WIN32_FIND_DATA FileData;
  HANDLE File = FindFirstFile(filter, &FileData);
  if (File != INVALID_HANDLE_VALUE)
  {
    bool Done = false;
    while (!Done)
    {
      if (FileData.dwFileAttributes != FILE_ATTRIBUTE_DIRECTORY)
      {
        char TempString[256];
        sprintf(TempString, "%s%s", dir, FileData.cFileName);
        Playlist.Add(TempString);
      }

      if (FindNextFile(File, &FileData) == false)
        Done = true;
    }

    FindClose(File);
  }

  if (Recurse)
  {
    // Recurses in directory if value is true
    File = FindFirstFile("*.*", &FileData);
    if (File == INVALID_HANDLE_VALUE)
      { SetCurrentDirectory(TempDir); return; }

    bool Done = false;
    while (!Done)
    {
      if (FileData.dwFileAttributes == FILE_ATTRIBUTE_DIRECTORY)
      {
        if (FileData.cFileName[0] != '.')
        {
          char TempString[256];
          sprintf(TempString, "%s%s\\", dir, FileData.cFileName);
          AddPlaylistDirectory(filter, TempString, Recurse);
        }
      }

      if (FindNextFile(File, &FileData) == false)
        Done = true;
    }

    FindClose(File);
  }

  SetCurrentDirectory(TempDir);
}


#pragma argsused
BOOL CALLBACK PlaylistDialogProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
  char FileName[256];
  OPENFILENAME ofn;

  switch (msg)
  {
    case WM_INITDIALOG:
      UpdatePlaylistDialog(hwnd);
      return 1;

    case WM_COMMAND:
      if (HIWORD(wparam) == LBN_DBLCLK)
        wparam = IDOK;

      switch (LOWORD(wparam))
      {
        case IDOK:
          int song;
          song = SendDlgItemMessage(hwnd, IDC_PLAYLISTBOX, LB_GETCURSEL, 0, 0);
          if (song == LB_ERR) song = -1;
          EndDialog(hwnd, song);
          return 1;

        case IDCANCEL: EndDialog(hwnd, -2); return 1;

        case IDC_ADD:
          memset(&ofn, 0, sizeof(ofn));
          ofn.lStructSize = sizeof(ofn);
          ofn.hwndOwner = hwnd;
          ofn.hInstance = hInst;
          ofn.lpstrFilter = "MIDI Files\0*.midi;*.mid\0All Files\0*.*\0";
          memset(FileName, 0, sizeof(FileName));
          ofn.lpstrFile = FileName;
          ofn.nMaxFile = 250;
          ofn.lpstrTitle = "Open MIDI File";
          ofn.Flags = OFN_EXPLORER | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_PATHMUSTEXIST;
          ofn.lpstrDefExt = "midi";

          if (GetOpenFileName(&ofn) == false) return 0;

          Playlist.Add(ofn.lpstrFile);
          SendDlgItemMessage(hwnd, IDC_PLAYLISTBOX, LB_ADDSTRING, 0, (LPARAM)ofn.lpstrFile);
          return 1;

        case IDC_ADDDIR:
          char Dir[256];

          if (GetDirDialog(hwnd, hInst, Dir) == 0)
            return 1;

          bool Recurse;
          if (MessageBox(hwnd, "Include subfolders?", AppName, MB_YESNO) == IDYES)
            Recurse = true;
          else
            Recurse = false;

          AddPlaylistDirectory("*.midi", Dir, Recurse);
          AddPlaylistDirectory("*.mid", Dir, Recurse);
          UpdatePlaylistDialog(hwnd);
          return 1;

        case IDC_REMOVE:
          char TempStr[256];
          int ItemIndex;
          ItemIndex = SendDlgItemMessage(hwnd, IDC_PLAYLISTBOX, LB_GETCURSEL, 0, 0);

          SendDlgItemMessage(hwnd, IDC_PLAYLISTBOX, LB_GETTEXT, ItemIndex, (LPARAM)TempStr);
          Playlist.Remove(TempStr);

          SendDlgItemMessage(hwnd, IDC_PLAYLISTBOX, LB_DELETESTRING, ItemIndex, 0);
          SendDlgItemMessage(hwnd, IDC_PLAYLISTBOX, LB_SETCURSEL, ItemIndex, 0);
          return 1;

        case IDC_REMOVEALL:
          Playlist.Clear();
          UpdatePlaylistDialog(hwnd);
          return 1;

        case IDC_SAVE:
          memset(&ofn, 0, sizeof(ofn));
          ofn.lStructSize = sizeof(ofn);
          ofn.hwndOwner = hwnd;
          ofn.hInstance = hInst;
          ofn.lpstrFilter = "Midnight Playlists\0*.mpl\0All Files\0*.*\0";
          memset(FileName, 0, sizeof(FileName));
          ofn.lpstrFile = FileName;
          ofn.nMaxFile = 250;
          ofn.lpstrTitle = "Save Playlist";
          ofn.Flags = OFN_EXPLORER | OFN_HIDEREADONLY | OFN_PATHMUSTEXIST;
          ofn.lpstrDefExt = "mpl";

          if (GetSaveFileName(&ofn) == true)
            Playlist.Save(FileName);
          return 1;

        case IDC_LOAD:
          OPENFILENAME ofn;
          memset(&ofn, 0, sizeof(ofn));
          ofn.lStructSize = sizeof(ofn);
          ofn.hwndOwner = hwnd;
          ofn.hInstance = hInst;
          ofn.lpstrFilter = "Midnight Playlists\0*.mpl\0All Files\0*.*\0";
          memset(FileName, 0, sizeof(FileName));
          ofn.lpstrFile = FileName;
          ofn.nMaxFile = 250;
          ofn.lpstrTitle = "Load Playlist";
          ofn.Flags = OFN_EXPLORER | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_PATHMUSTEXIST;
          ofn.lpstrDefExt = "mpl";

          if (GetOpenFileName(&ofn) == true)
          {
            Playlist.Load(FileName);
            UpdatePlaylistDialog(hwnd);
          }
          return 1;
      }
      return 0;


    default: return 0;
  }
}


int PlaylistDialog()
{
  return DialogBox(hInst, MAKEINTRESOURCE(IDD_PLAYLIST), MainWindow, (DLGPROC)PlaylistDialogProc);
}


bool PLAYLIST::Clear()
{
  if (FileNames != NULL)
    delete FileNames;
  FileNames = NULL;
  NumFiles = 0;
  CurFile = -1;
  return true;
}


bool PLAYLIST::Add(char *filename)
{
  SFILENAME *OldFileNames = FileNames;
  FileNames = new SFILENAME[NumFiles + 1];
  if (FileNames == NULL) { FileNames = OldFileNames; return false; }

  NumFiles++;

  if (NumFiles == 1)
    { strcpy(FileNames[0], filename); return true; }

  bool Flag = true;
  int i, j = 0;
  for (i = 0; i < NumFiles; i++)
  {
    strcpy(FileNames[i], OldFileNames[j]);
    if (Flag && strcmp(filename, FileNames[j]) < 0)
    {
      strcpy(FileNames[i], filename);
      j--;
      Flag = false;
    }

    j++;

    if (Flag && j == NumFiles)
      strcpy(FileNames[i], filename);
  }

  delete OldFileNames;
  return true;
}


int PLAYLIST::InList(char *filename)
{
  int i;
  for (i = 0; i < NumFiles; i++)
  {
    if (strcmp(filename, FileNames[i]) == 0)
      return i;
  }
  return -1;
}


bool PLAYLIST::Remove(char *filename)
{
  int ItemNum = InList(filename);
  if (ItemNum != -1)
  {
    SFILENAME *OldFiles = FileNames;
    FileNames = new SFILENAME[NumFiles - 1];
    if (FileNames == NULL) { FileNames == OldFiles; return false; }

    int i;
    for (i = 0; i < ItemNum; i++)
      strcpy(FileNames[i], OldFiles[i]);

    for (i = ItemNum; i < NumFiles - 1; i++)
      strcpy(FileNames[i], OldFiles[i + 1]);

    if (CurFile > ItemNum) CurFile--;  

    NumFiles--;
    delete OldFiles;

    return true;
  }
  return false;
}


bool PLAYLIST::Save(char *filename)
{
  FILE *fstrm = fopen(filename, "wb");
  if (fstrm == NULL) return false;

  long NumSongs = NumFiles;
  fwrite(&NumSongs, 1, sizeof(NumSongs), fstrm);

  long i;
  for (i = 0; i < NumSongs; i++)
  {
    fwrite(FileNames[i], 1, sizeof(SFILENAME), fstrm);
  }

  fclose(fstrm);
  return true;
}


bool PLAYLIST::Load(char *filename)
{
  FILE *fstrm = fopen(filename, "rb");
  if (fstrm == NULL) return false;
  Clear();

  long NumSongs;
  fread(&NumSongs, 1, sizeof(NumSongs), fstrm);

  SFILENAME TempFile;

  long i;
  for (i = 0; i < NumSongs; i++)
  {
    fread(TempFile, 1, sizeof(SFILENAME), fstrm);
    Add(TempFile);
  }

  fclose(fstrm);
  return true;
}


char *PLAYLIST::GetCurrentSong()
{
  if (CurFile == -1) return NULL;
  return FileNames[CurFile];
}


int PLAYLIST::GetCurrentIndex()
{
  return CurFile;
}


bool PLAYLIST::SetCurrentSong(int song)
{
  if (song < NumFiles && song >= 0)
    { CurFile = song; return true; }
  return false;
}


bool PLAYLIST::Next()
{
  if (NumFiles > 0)
  {
    if (Shuffle)
      CurFile = rand() % NumFiles;
    else
    {
      CurFile++;
      if (CurFile >= NumFiles)
        CurFile = 0;
    }
    return true;
  }
  return false;
}


bool PLAYLIST::Prev()
{
  if (NumFiles > 0)
  {
    if (Shuffle)
      CurFile = rand() % NumFiles;
    else
    {
      CurFile--;
      if (CurFile < 0)
        CurFile = NumFiles - 1;
    }
    return true;
  }
  return false;
}


int PLAYLIST::GetNumFiles()
{
  return NumFiles;
}


SFILENAME *PLAYLIST::GetValue(int index)
{
  return &FileNames[index];
}
